@@ -200,6 +200,33 @@ static gboolean start_another_request (GIOChannel    *serv_chan,
 				       GIOCondition   condition, 
 				       gpointer       user_data);
 
+static gboolean
+check_close_connection (SoupMessage *msg)
+{
+	const char *connection_hdr;
+	gboolean close_connection;
+
+	connection_hdr = soup_message_get_header (msg->request_headers,
+						  "Connection");
+
+	if (msg->priv->http_version == SOUP_HTTP_1_0) {
+		if (connection_hdr && g_strcasecmp (connection_hdr,
+						    "keep-alive") == 0)
+			close_connection = FALSE;
+		else
+			close_connection = TRUE;
+	}
+	else {
+		if (connection_hdr && g_strcasecmp (connection_hdr,
+						    "close") == 0)
+			close_connection = TRUE;
+		else
+			close_connection = FALSE;
+	}
+
+	return close_connection;
+} /* check_close_connection */
+
 static void
 destroy_message (SoupMessage *msg)
 {
@@ -208,18 +235,24 @@ destroy_message (SoupMessage *msg)
 	SoupServerMessage *server_msg = msg->priv->server_msg;
 
 	if (server_sock) {
-		if (server_msg && msg->priv->http_version == SOUP_HTTP_1_0)
-			/*
-			 * Close the socket if we are using HTTP/1.0 and 
-			 * did not specify a Content-Length response header.
-			 */
+		GIOChannel *chan;
+
+		chan = soup_socket_get_iochannel (server_sock);
+
+		/*
+		 * Close the socket if we're using HTTP/1.0 and
+		 * "Connection: keep-alive" isn't specified, or if we're
+		 * using HTTP/1.1 and "Connection: close" was specified.
+		 */
+		if (check_close_connection (msg)) {
+			g_io_channel_close (chan);
 			soup_socket_unref (server_sock);
+		}
 		else {
 			/*
 			 * Listen for another request on this connection
 			 */
 			ServerConnectData *data;
-			GIOChannel *chan;
 
 			data = g_new0 (ServerConnectData, 1);
 			data->server = msg->priv->server;
@@ -426,7 +459,7 @@ get_response_header (SoupMessage          *req,
 {
 	GString *ret = g_string_new (NULL);
 
-	if (status_line) 
+	if (status_line)
 		g_string_sprintfa (ret, 
 				   "HTTP/1.1 %d %s\r\n", 
 				   req->errorcode, 
@@ -883,6 +916,7 @@ message_new (SoupServer *server)
 	 */
 	msg = soup_message_new (NULL, NULL);
 	if (msg) {
+		msg->method = NULL;
 		msg->priv->server = server;
 		soup_server_ref (server);
 	}
@@ -949,7 +983,7 @@ conn_accept (GIOChannel    *serv_chan,
 	chan = soup_socket_get_iochannel (sock);
 
 	if (server->proto == SOUP_PROTOCOL_HTTPS) {
-		chan = soup_ssl_get_iochannel (chan);
+		chan = soup_ssl_get_server_iochannel (chan);
 		g_io_channel_unref (sock->iochannel);
 		g_io_channel_ref (chan);
 		sock->iochannel = chan;
@@ -1179,6 +1213,33 @@ soup_server_get_handler (SoupServer *server, const gchar *path)
 	return server->default_handler;
 }
 
+SoupAddress *
+soup_server_context_get_client_address (SoupServerContext *context)
+{
+	SoupSocket *socket;
+	SoupAddress *address;
+
+	g_return_val_if_fail (context != NULL, NULL);
+
+	socket = context->msg->priv->server_sock;
+	address = soup_socket_get_address (socket);
+
+	return address;
+}
+
+gchar *
+soup_server_context_get_client_host (SoupServerContext *context)
+{
+	gchar *host;
+	SoupAddress *address;
+
+	address = soup_server_context_get_client_address (context);
+	host = g_strdup (soup_address_get_canonical_name (address));
+	soup_address_unref (address);
+	
+	return host;
+}
+
 static SoupServerAuthContext *
 auth_context_copy (SoupServerAuthContext *auth_ctx)
 {
